/* 
 *
 * $Id: k3bmonkeyencoder.cpp,v 1.3 2004/11/11 18:58:13 trueg Exp $
 * Copyright (C) 2003 Sebastian Trueg <trueg@k3b.org>
 *
 * This file is part of the K3b project.
 * Copyright (C) 1998-2003 Sebastian Trueg <trueg@k3b.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * See the file "COPYING" for the exact licensing terms.
 */

#include "k3bmonkeyencoder.h"
#include "k3bmonkeyio.h"

#include <All.h>
#include <MACLib.h>
#include <APETag.h>

#include <k3bcore.h>
#include <k3bpluginfactory.h>

#include <klocale.h>
#include <kconfig.h>
#include <kdebug.h>
#include <kinstance.h>

#include <qlayout.h>
#include <qcstring.h>
#include <qfile.h>
#include <qlabel.h>
#include <qcombobox.h>
#include <qpair.h>
#include <qvaluelist.h>


K_EXPORT_COMPONENT_FACTORY( libk3bmonkeyencoder, K3bPluginFactory<K3bMonkeyEncoder>( "k3bmonkeyplugin" ) )


static int compressionLevelFromString( const QString& s )
{
  if( s == "fast" )
    return COMPRESSION_LEVEL_FAST;
  else if( s == "high" )
    return COMPRESSION_LEVEL_HIGH;
  else if( s == "extra high" )
    return COMPRESSION_LEVEL_EXTRA_HIGH;
//   else if( s == "insane high" )
//     return COMPRESSION_LEVEL_INSANE_HIGH;
//   else if( s == "braindead high" )
//     return COMPRESSION_LEVEL_BRAINDEAD_HIGH;
  else
    return COMPRESSION_LEVEL_NORMAL;
}



class K3bMonkeyEncoder::Private
{
public:
  Private() 
    : started(false),
      encoder(0),
      io(0) {
  }

  QString filename;
  bool started;
  IAPECompress* encoder;
  K3bMonkeyIO* io;

  QValueList< QPair<const str_utf16*, QString> > tags;
};


K3bMonkeyEncoder::K3bMonkeyEncoder( QObject* parent, const char* name )
  : K3bAudioEncoder( parent, name )
{
  d = new Private();
}


K3bMonkeyEncoder::~K3bMonkeyEncoder()
{
  delete d->encoder;
  delete d->io;
  delete d;
}


bool K3bMonkeyEncoder::openFile( const QString&, const QString& filename, const K3b::Msf& length )
{
  kdDebug() << "(K3bMonkeyEncoder) opening file " << filename << endl;

  d->tags.clear();

  if( !d->encoder )
    d->encoder = CreateIAPECompress();

  if( !d->io )
    d->io = new K3bMonkeyIO();

  if( !d->encoder ) {
    kdDebug() << "(K3bMonkeyEncoder) failed to create Monkey encoder." << endl;
    return false;
  }

  d->filename = filename;

  KConfig* c = k3bcore->config();
  c->setGroup( "K3bMonkeyEncoderPlugin" );
  int level = compressionLevelFromString(c->readEntry( "compression level", "normal" ));

  d->io->Open( filename );

  WAVEFORMATEX wfeAudioFormat;
  FillWaveFormatEx(&wfeAudioFormat, 44100, 16, 2);
  d->started = (d->encoder->StartEx( d->io, 
				     &wfeAudioFormat,
				     length.audioBytes(), //FIXME: is this correct???
				     level,
				     0,
				     CREATE_WAV_HEADER_ON_DECOMPRESSION ) == 0 );

  return d->started;
}


bool K3bMonkeyEncoder::isOpen() const
{
  return d->started;
}


void K3bMonkeyEncoder::closeFile()
{
  if( d->started ) {
    if( d->encoder->Finish( 0, 0, 0 ) ) {
      kdDebug() << "(K3bMonkeyEncoder) error while finishing encoding." << endl;
    }
    
    if( !d->tags.isEmpty() ) {
      CAPETag tag( d->io );
      for( QValueList< QPair<const str_utf16*, QString> >::iterator it = d->tags.begin();
	   it != d->tags.end(); ++it )
	tag.SetFieldString( (*it).first, 
			    GetUTF16FromUTF8( (const str_utf8*)(*it).second.utf8().data() ) );
      tag.Save();
    }

    d->io->Close();

    d->started = false;
  }
}


const QString& K3bMonkeyEncoder::filename() const
{
  return d->filename;
}


long K3bMonkeyEncoder::encodeInternal( const char* data, Q_ULONG len )
{
  if( d->started ) {
    int bufSize;
    // lock the compression buffer
    unsigned char* buffer = d->encoder->LockBuffer( &bufSize );

    // copy data into the compression buffer
    int count = static_cast<int>( QMIN( len, static_cast<Q_ULONG>(bufSize) ) );
    for( int i = 0; i < count; ++i ) {
      buffer[i] = data[i];
    }

    // unlock the buffer and let it get processed
    if( d->encoder->UnlockBuffer( count, true ) != 0 )
      return -1;
    else if( count < len )
      return count + encodeInternal( data+count, len - count );
    else
      return count;
  }
  else
    return -1;
}


void K3bMonkeyEncoder::setMetaDataInternal( K3bAudioEncoder::MetaDataField f, const QString& value )
{
  const str_utf16* key = 0;

  switch( f ) {
  case META_TRACK_TITLE:
    key = APE_TAG_FIELD_TITLE;
    break;
  case META_TRACK_ARTIST:
    key = APE_TAG_FIELD_ARTIST;
    break;
  case META_TRACK_COMMENT:
    key = APE_TAG_FIELD_COMMENT;
    break;
  case META_ALBUM_TITLE:
    key = APE_TAG_FIELD_ALBUM;
    break;
  case META_YEAR:
    key = APE_TAG_FIELD_YEAR;
    break;
  case META_TRACK_NUMBER:
    key = APE_TAG_FIELD_TRACK;
    break;
  case META_GENRE:
    key = APE_TAG_FIELD_GENRE;
    break;
  default:
    return;
  }

  if( key )
    d->tags.append( qMakePair( key, value ) );
}





K3bMonkeyEncoderSettingsWidget::K3bMonkeyEncoderSettingsWidget( QWidget* parent, const char* name )
  : K3bPluginConfigWidget( parent, name )
{
  QHBoxLayout* layout = new QHBoxLayout( this );
  layout->setAutoAdd( true );
  (void)new QLabel( i18n("Compression level:"), this );
  m_comboCompressionLevel = new QComboBox( this );

  m_comboCompressionLevel->insertItem( i18n("Fast") );
  m_comboCompressionLevel->insertItem( i18n("Normal") );
  m_comboCompressionLevel->insertItem( i18n("High") );
  m_comboCompressionLevel->insertItem( i18n("Extra High") );
//   m_comboCompressionLevel->insertItem( i18n("Insane High") );
//   m_comboCompressionLevel->insertItem( i18n("Braindead High") );
}


K3bMonkeyEncoderSettingsWidget::~K3bMonkeyEncoderSettingsWidget()
{
}


void K3bMonkeyEncoderSettingsWidget::saveConfig()
{
  KConfig* c = k3bcore->config();
  c->setGroup( "K3bMonkeyEncoderPlugin" );

  QString s = "normal";
  switch( m_comboCompressionLevel->currentItem() ) {
  case 0:
    s = "fast";
    break;
  case 1:
    s = "normal";
    break;
  case 2:
    s = "high";
    break;
  case 3:
    s = "extra high";
    break;
//   case 4:
//     s = "insane high";
//     break;
//   case 5:
//     s = "braindead high";
//     break;
  }

  c->writeEntry( "compression level", s );
}


void K3bMonkeyEncoderSettingsWidget::loadConfig()
{
  KConfig* c = k3bcore->config();
  c->setGroup( "K3bMonkeyEncoderPlugin" );

  switch( compressionLevelFromString(c->readEntry( "compression level", "normal" )) ) {
  case COMPRESSION_LEVEL_FAST:
    m_comboCompressionLevel->setCurrentItem( 0 );
    break;
  case COMPRESSION_LEVEL_NORMAL:
    m_comboCompressionLevel->setCurrentItem( 1 );
    break;
  case COMPRESSION_LEVEL_HIGH:
    m_comboCompressionLevel->setCurrentItem( 2 );
    break;
  case COMPRESSION_LEVEL_EXTRA_HIGH:
    m_comboCompressionLevel->setCurrentItem( 3 );
    break;
//   case COMPRESSION_LEVEL_INSANE_HIGH:
//     m_comboCompressionLevel->setCurrentItem( 4 );
//     break;
//   case COMPRESSION_LEVEL_BRAINDEAD_HIGH:
//     m_comboCompressionLevel->setCurrentItem( 5 );
//     break;
  }
}


QStringList K3bMonkeyEncoder::extensions() const
{
  return QStringList( "ape" );
}


QString K3bMonkeyEncoder::fileTypeComment( const QString& ) const
{
  return i18n("Monkey Audio");
}


long long K3bMonkeyEncoder::fileSize( const QString&, const K3b::Msf& msf ) const
{
  // FIXME
  return -1;
}


K3bPluginConfigWidget* K3bMonkeyEncoder::createConfigWidget( QWidget* parent, 
								    const char* name ) const
{
  return new K3bMonkeyEncoderSettingsWidget( parent, name );
}


#include "k3bmonkeyencoder.moc"
